import json
import logging
import socket

from django.apps import apps as django_apps
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http.response import HttpResponse
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.views.generic.base import TemplateView
from edc_base.view_mixins import EdcBaseViewMixin
from edc_navbar import NavbarViewMixin
from django_collect_offline_files.action_handler import ActionHandler, ActionHandlerError

from ..admin import django_collect_offline_admin
from ..offline_view_mixin import OfflineViewMixin
from ..site_offline_models import site_offline_models

app_config = django_apps.get_app_config('django_collect_offline_files')
logger = logging.getLogger('django_collect_offline')


@method_decorator(never_cache, name='dispatch')
@method_decorator(login_required, name='dispatch')
class HomeView(EdcBaseViewMixin, NavbarViewMixin, OfflineViewMixin, TemplateView):

    template_name = 'django_collect_offline/home.html'
    action_handler_cls = ActionHandler

    navbar_name = 'django_collect_offline'
    navbar_selected_item = 'synchronization'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._action_handler = None

    @property
    def action_handler(self):
        if not self._action_handler:
            self._action_handler = self.action_handler_cls(
                src_path=app_config.outgoing_folder,
                dst_tmp=app_config.tmp_folder,
                dst_path=app_config.incoming_folder,
                archive_path=app_config.archive_folder,
                username=app_config.user,
                remote_host=app_config.remote_host)
        return self._action_handler

    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        app_config = django_apps.get_app_config('django_collect_offline')
        context.update(
            base_template_name=app_config.base_template_name,
            cors_origin_whitelist=self.cors_origin_whitelist,
            django_collect_offline_admin=django_collect_offline_admin,
            django_collect_offline_app_config=app_config,
            django_collect_offline_files_app_config=django_apps.get_app_config(
                'django_collect_offline_files'),
            django_collect_offline_role=self.device_role,
            hostname=socket.gethostname(),
            ip_address=django_apps.get_app_config(
                'django_collect_offline_files').remote_host,
            pending_files=self.action_handler.pending_filenames,
            recently_sent_files=self.action_handler.sent_history[0:20],
            site_models=site_offline_models.site_models)
        return context

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        if request.is_ajax():
            action = request.GET.get('action')
            try:
                self.action_handler.action(label=action)
            except ActionHandlerError as e:
                logger.warn(e)
                logger.exception(e)
                response_data = dict(errmsg=str(e))
            else:
                response_data = self.action_handler.data
            return HttpResponse(
                json.dumps(response_data), content_type='application/json')
        return self.render_to_response(context)

    @property
    def cors_origin_whitelist(self):
        try:
            cors_origin_whitelist = settings.CORS_ORIGIN_WHITELIST
        except AttributeError:
            cors_origin_whitelist = []
        return cors_origin_whitelist



import logging

from django.apps import apps as django_apps
from django.core.management.base import BaseCommand

from ...transaction import CustomTransactionDeserializer


app_config = django_apps.get_app_config('django_collect_offline')
logger = logging.getLogger('django_collect_offline')


class Command(BaseCommand):
    """Usage:
        python manage.py deserialize --batch=9835201711152020
            --model=label_lower --order_by=created,producer
    """

    help = ('Deserialises transactions manually using '
            'different filter options.')

    def add_arguments(self, parser):
        parser.add_argument(
            '--model',
            dest='model',
            default=None,
            help=('Specify the model name/tx name.'),
        )

        parser.add_argument(
            '--batch',
            dest='batch',
            default=None,
            help=('Specify batch.'),
        )

        parser.add_argument(
            '--order_by',
            dest='order_by',
            default='created',
            help=('Specify a field to order by e.g timestamp or created.'),
        )

        parser.add_argument(
            '--producer',
            dest='producer',
            default=None,
            help=('Specify a producer/client machine. e.g bcpp010'),
        )

    def handle(self, *args, **options):
        CustomTransactionDeserializer(**options)



from django.core import serializers


def deserialize(json_text=None):
    """Returns a generator of deserialized objects.

    Wraps django deserialize with defaults for JSON
    and natural keys.

    See https://docs.djangoproject.com/en/2.1/topics/serialization/
    """

    return serializers.deserialize(
        "json", json_text,
        ensure_ascii=True,
        use_natural_foreign_keys=True,
        use_natural_primary_keys=True)



import sys

from django.apps import AppConfig as DjangoAppConfig
from django.core.management.color import color_style

from .site_offline_models import site_offline_models

style = color_style()


class OfflineConfigError(Exception):
    pass


class AppConfig(DjangoAppConfig):
    name = 'django_collect_offline'
    verbose_name = 'Offline Synchronization'
    base_template_name = 'edc_dashboard/bootstrap3/base.html'
    custom_json_parsers = []
    django_collect_offline_files_using = True

    # see edc_device for ROLE

    def ready(self):
        from .signals import (
            create_auth_token, serialize_on_post_delete,
            serialize_m2m_on_save, serialize_on_save)
        sys.stdout.write(f'Loading {self.verbose_name} ...\n')
        site_offline_models.autodiscover()
        sys.stdout.write(f' Done loading {self.verbose_name}.\n')



from django.contrib.admin.sites import AdminSite


class DjangoOfflineAdminSite(AdminSite):
    site_header = 'Django Offline'
    site_title = 'Django Offline'
    index_title = 'Django Offline Administration'
    site_url = '/administration/'


django_collect_offline_admin = DjangoOfflineAdminSite(name='django_collect_offline_admin')



from django.core.exceptions import MultipleObjectsReturned
from django.test import TestCase, tag
from django.test.utils import override_settings

from ..constants import INSERT, UPDATE
from ..models import OutgoingTransaction
from ..site_offline_models import site_offline_models
from ..offline_model import OfflineHistoricalManagerError, OfflineUuidPrimaryKeyMissing
from ..offline_model import OfflineModel
from ..offline_model import OfflineNaturalKeyMissing, OfflineGetByNaturalKeyMissing
from .models import TestModel, BadTestModel, AnotherBadTestModel, YetAnotherBadTestModel
from .models import TestModelWithFkProtected
from .models import TestOfflineModelNoHistoryManager, TestOfflineModelNoUuid


class TestOffline(TestCase):

    multi_db = True

    def setUp(self):
        site_offline_models.registry = {}
        site_offline_models.loaded = False
        offline_models = ['django_collect_offline.testmodel',
                          'django_collect_offline.badtestmodel',
                          'django_collect_offline.anotherbadtestmodel',
                          'django_collect_offline.yetanotherbadtestmodel',
                          'django_collect_offline.testmodelwithfkprotected',
                          'django_collect_offline.testmodelwithm2m',
                          'django_collect_offline.testofflinemodelnohistorymanager',
                          'django_collect_offline.testofflinemodelnouuid']
        site_offline_models.register(models=offline_models)

    def get_credentials(self):
        return self.create_apikey(username=self.username,
                                  api_key=self.api_client_key)

    def test_str(self):
        obj = TestModel()
        obj = OfflineModel(obj)
        self.assertTrue(str(obj))
        self.assertTrue(repr(obj))

    def test_raises_on_missing_natural_key(self):
        with override_settings(DEVICE_ID='10'):
            with self.assertRaises(OfflineNaturalKeyMissing):
                BadTestModel.objects.using('client').create()

    def test_raises_on_missing_get_by_natural_key(self):
        with override_settings(DEVICE_ID='10'):
            with self.assertRaises(OfflineGetByNaturalKeyMissing):
                AnotherBadTestModel.objects.using('client').create()

    def test_raises_on_wrong_type_of_historical_manager(self):
        with override_settings(DEVICE_ID='10'):
            with self.assertRaises(OfflineHistoricalManagerError):
                YetAnotherBadTestModel.objects.using('client').create()

    def test_raises_on_no_historical_manager(self):
        with override_settings(DEVICE_ID='10'):
            try:
                TestOfflineModelNoHistoryManager.objects.using(
                    'client').create()
            except OfflineHistoricalManagerError:
                self.fail('OfflineHistoricalManagerError unexpectedly raised.')

    @tag('2')
    def test_raises_on_missing_uuid_primary_key(self):
        with override_settings(DEVICE_ID='10'):
            with self.assertRaises(OfflineUuidPrimaryKeyMissing):
                TestOfflineModelNoUuid.objects.using('client').create()

    def test_creates_outgoing_on_add(self):
        with override_settings(DEVICE_ID='10'):
            test_model = TestModel.objects.using('client').create(f1='erik')
            with self.assertRaises(OutgoingTransaction.DoesNotExist):
                try:
                    OutgoingTransaction.objects.using('client').get(
                        tx_pk=test_model.pk,
                        tx_name='django_collect_offline.testmodel',
                        action=INSERT)
                except OutgoingTransaction.DoesNotExist:
                    pass
                else:
                    raise OutgoingTransaction.DoesNotExist()
            history_obj = test_model.history.using(
                'client').get(id=test_model.id)
            with self.assertRaises(OutgoingTransaction.DoesNotExist):
                try:
                    OutgoingTransaction.objects.using('client').get(
                        tx_pk=history_obj.history_id,
                        tx_name='django_collect_offline.historicaltestmodel',
                        action=INSERT)
                except OutgoingTransaction.DoesNotExist:
                    pass
                else:
                    raise OutgoingTransaction.DoesNotExist()

    @tag('1')
    def test_creates_outgoing_on_add_with_fk_in_order(self):
        with override_settings(DEVICE_ID='10'):
            outgoing = {}
            test_model = TestModel.objects.using('client').create(f1='erik')
            test_model_with_fk = TestModelWithFkProtected.objects.using(
                'client').create(f1='f1', test_model=test_model)
            outgoing.update(
                test_model=OutgoingTransaction.objects.using('client').get(
                    tx_pk=test_model.pk,
                    tx_name='django_collect_offline.testmodel',
                    action=INSERT))
            history_obj = test_model.history.using(
                'client').get(id=test_model.id)
            outgoing.update(
                test_model_historical=OutgoingTransaction.objects.using('client').get(
                    tx_pk=history_obj.history_id,
                    tx_name='django_collect_offline.historicaltestmodel',
                    action=INSERT))

            with self.assertRaises(OutgoingTransaction.DoesNotExist):
                try:
                    outgoing.update(
                        test_model_with_fk=OutgoingTransaction.objects.using('client').get(
                            tx_pk=test_model_with_fk.pk,
                            tx_name='django_collect_offline.testmodelwithfkprotected',
                            action=INSERT))
                except OutgoingTransaction.DoesNotExist:
                    pass
                else:
                    raise OutgoingTransaction.DoesNotExist()
            history_obj = test_model_with_fk.history.using(
                'client').get(id=test_model_with_fk.id)
            with self.assertRaises(OutgoingTransaction.DoesNotExist):
                try:
                    outgoing.update(
                        test_model_with_fk_historical=(
                            OutgoingTransaction.objects.using('client').get(
                                tx_pk=history_obj.history_id,
                                tx_name=(
                                    'django_collect_offline.historicaltest'
                                    'modelwithfkprotected'),
                                action=INSERT)))
                except OutgoingTransaction.DoesNotExist:
                    pass
                else:
                    raise OutgoingTransaction.DoesNotExist()

    @override_settings(ALLOW_MODEL_SERIALIZATION=False)
    def test_does_not_create_outgoing(self):
        with override_settings(DEVICE_ID='10', ALLOW_MODEL_SERIALIZATION=False):
            test_model = TestModel.objects.using('client').create(f1='erik')
            with self.assertRaises(OutgoingTransaction.DoesNotExist):
                OutgoingTransaction.objects.using(
                    'client').get(tx_pk=test_model.pk)

    def test_creates_outgoing_on_change(self):
        with override_settings(DEVICE_ID='10'):
            test_model = TestModel.objects.using('client').create(f1='erik')
            test_model.save(using='client')
            with self.assertRaises(OutgoingTransaction.DoesNotExist):
                try:
                    OutgoingTransaction.objects.using('client').get(
                        tx_pk=test_model.pk,
                        tx_name='django_collect_offline.testmodel',
                        action=INSERT)
                    OutgoingTransaction.objects.using('client').get(
                        tx_pk=test_model.pk,
                        tx_name='django_collect_offline.testmodel',
                        action=UPDATE)
                except OutgoingTransaction.DoesNotExist:
                    pass
                else:
                    raise OutgoingTransaction.DoesNotExist()
            self.assertEqual(
                2, OutgoingTransaction.objects.using('client').filter(
                    tx_name='django_collect_offline.historicaltestmodel',
                    action=INSERT).count())

    def test_timestamp_is_default_order(self):
        with override_settings(DEVICE_ID='10'):
            test_model = TestModel.objects.using('client').create(f1='erik')
            test_model.save(using='client')
            last = 0
            for obj in OutgoingTransaction.objects.using('client').all():
                self.assertGreaterEqual(int(obj.timestamp), last)
                last = int(obj.timestamp)

    def test_created_obj_serializes_to_correct_db(self):
        """Asserts that the obj and the audit obj serialize to the
        correct DB in a multi-database environment.
        """
        TestModel.objects.using('client').create(f1='erik')
        result = [
            obj.tx_name for obj in OutgoingTransaction.objects.using('client').all()]
        result.sort()
        self.assertListEqual(
            result,
            ['django_collect_offline.historicaltestmodel', 'django_collect_offline.testmodel'])
        self.assertListEqual(
            [obj.tx_name for obj in OutgoingTransaction.objects.using('server').all()], [])
        self.assertRaises(
            OutgoingTransaction.DoesNotExist,
            OutgoingTransaction.objects.using('server').get,
            tx_name='django_collect_offline.testmodel')
        self.assertRaises(
            MultipleObjectsReturned,
            OutgoingTransaction.objects.using('client').get,
            tx_name__contains='testmodel')

    def test_updated_obj_serializes_to_correct_db(self):
        """Asserts that the obj and the audit obj serialize to the
        correct DB in a multi-database environment.
        """
        test_model = TestModel.objects.using('client').create(f1='erik')
        result = [obj.tx_name for obj in OutgoingTransaction.objects.using(
            'client').filter(action=INSERT)]
        result.sort()
        self.assertListEqual(
            result, ['django_collect_offline.historicaltestmodel',
                     'django_collect_offline.testmodel'])
        self.assertListEqual(
            [obj.tx_name for obj in OutgoingTransaction.objects.using(
                'client').filter(action=UPDATE)],
            [])
        test_model.save(using='client')
        self.assertListEqual(
            [obj.tx_name for obj in OutgoingTransaction.objects.using(
                'client').filter(action=UPDATE)],
            ['django_collect_offline.testmodel'])
        result = [obj.tx_name for obj in OutgoingTransaction.objects.using(
            'client').filter(action=INSERT)]
        result.sort()
        self.assertListEqual(
            result,
            ['django_collect_offline.historicaltestmodel',
             'django_collect_offline.historicaltestmodel',
             'django_collect_offline.testmodel'])



# Generated by Django 2.0.6 on 2018-07-03 10:45

import _socket
from django.db import migrations, models
import django.db.models.deletion
import django_revision.revision_field
import edc_base.sites.managers
import edc_base.utils
import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
import edc_model_fields.fields.uuid_auto_field


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
    ]

    operations = [
        migrations.CreateModel(
            name='Client',
            fields=[
                ('created', models.DateTimeField(blank=True, default=edc_base.utils.get_utcnow)),
                ('modified', models.DateTimeField(blank=True, default=edc_base.utils.get_utcnow)),
                ('user_created', edc_model_fields.fields.userfield.UserField(blank=True, help_text='Updated by admin.save_model', max_length=50, verbose_name='user created')),
                ('user_modified', edc_model_fields.fields.userfield.UserField(blank=True, help_text='Updated by admin.save_model', max_length=50, verbose_name='user modified')),
                ('hostname_created', models.CharField(blank=True, default=_socket.gethostname, help_text='System field. (modified on create only)', max_length=60)),
                ('hostname_modified', edc_model_fields.fields.hostname_modification_field.HostnameModificationField(blank=True, help_text='System field. (modified on every save)', max_length=50)),
                ('revision', django_revision.revision_field.RevisionField(blank=True, editable=False, help_text='System field. Git repository tag:branch:commit.', max_length=75, null=True, verbose_name='Revision')),
                ('device_created', models.CharField(blank=True, max_length=10)),
                ('device_modified', models.CharField(blank=True, max_length=10)),
                ('id', edc_model_fields.fields.uuid_auto_field.UUIDAutoField(blank=True, editable=False, help_text='System auto field. UUID primary key.', primary_key=True, serialize=False)),
                ('hostname', models.CharField(max_length=200, unique=True)),
                ('port', models.IntegerField(default='80')),
                ('api_name', models.CharField(default='v1', max_length=15)),
                ('format', models.CharField(default='json', max_length=15)),
                ('authentication', models.CharField(default='api_key', max_length=15)),
                ('is_active', models.BooleanField(default=True)),
                ('last_sync_datetime', models.DateTimeField(blank=True, null=True)),
                ('last_sync_status', models.CharField(blank=True, default='-', max_length=250, null=True)),
                ('comment', models.TextField(blank=True, max_length=50, null=True)),
            ],
            options={
                'ordering': ['hostname', 'port'],
            },
        ),
        migrations.CreateModel(
            name='History',
            fields=[
                ('created', models.DateTimeField(blank=True, default=edc_base.utils.get_utcnow)),
                ('modified', models.DateTimeField(blank=True, default=edc_base.utils.get_utcnow)),
                ('user_created', edc_model_fields.fields.userfield.UserField(blank=True, help_text='Updated by admin.save_model', max_length=50, verbose_name='user created')),
                ('user_modified', edc_model_fields.fields.userfield.UserField(blank=True, help_text='Updated by admin.save_model', max_length=50, verbose_name='user modified')),
                ('hostname_created', models.CharField(blank=True, default=_socket.gethostname, help_text='System field. (modified on create only)', max_length=60)),
                ('hostname_modified', edc_model_fields.fields.hostname_modification_field.HostnameModificationField(blank=True, help_text='System field. (modified on every save)', max_length=50)),
                ('revision', django_revision.revision_field.RevisionField(blank=True, editable=False, help_text='System field. Git repository tag:branch:commit.', max_length=75, null=True, verbose_name='Revision')),
                ('device_created', models.CharField(blank=True, max_length=10)),
                ('device_modified', models.CharField(blank=True, max_length=10)),
                ('id', edc_model_fields.fields.uuid_auto_field.UUIDAutoField(blank=True, editable=False, help_text='System auto field. UUID primary key.', primary_key=True, serialize=False)),
                ('filename', models.CharField(max_length=100, unique=True)),
                ('hostname', models.CharField(max_length=100)),
                ('sent_datetime', models.DateTimeField(default=edc_base.utils.get_utcnow)),
            ],
            options={
                'verbose_name': 'Sent History',
                'verbose_name_plural': 'Sent History',
                'ordering': ('-sent_datetime',),
            },
        ),
        migrations.CreateModel(
            name='IncomingTransaction',
            fields=[
                ('created', models.DateTimeField(blank=True, default=edc_base.utils.get_utcnow)),
                ('modified', models.DateTimeField(blank=True, default=edc_base.utils.get_utcnow)),
                ('user_created', edc_model_fields.fields.userfield.UserField(blank=True, help_text='Updated by admin.save_model', max_length=50, verbose_name='user created')),
                ('user_modified', edc_model_fields.fields.userfield.UserField(blank=True, help_text='Updated by admin.save_model', max_length=50, verbose_name='user modified')),
                ('hostname_created', models.CharField(blank=True, default=_socket.gethostname, help_text='System field. (modified on create only)', max_length=60)),
                ('hostname_modified', edc_model_fields.fields.hostname_modification_field.HostnameModificationField(blank=True, help_text='System field. (modified on every save)', max_length=50)),
                ('revision', django_revision.revision_field.RevisionField(blank=True, editable=False, help_text='System field. Git repository tag:branch:commit.', max_length=75, null=True, verbose_name='Revision')),
                ('device_created', models.CharField(blank=True, max_length=10)),
                ('device_modified', models.CharField(blank=True, max_length=10)),
                ('id', edc_model_fields.fields.uuid_auto_field.UUIDAutoField(blank=True, editable=False, help_text='System auto field. UUID primary key.', primary_key=True, serialize=False)),
                ('tx', models.BinaryField()),
                ('tx_name', models.CharField(max_length=64)),
                ('tx_pk', models.UUIDField(db_index=True)),
                ('producer', models.CharField(db_index=True, help_text='Producer name', max_length=200)),
                ('action', models.CharField(choices=[('I', 'Insert'), ('U', 'Update'), ('D', 'Delete')], max_length=1)),
                ('timestamp', models.CharField(db_index=True, max_length=50)),
                ('consumed_datetime', models.DateTimeField(blank=True, null=True)),
                ('consumer', models.CharField(blank=True, max_length=200, null=True)),
                ('is_ignored', models.BooleanField(default=False)),
                ('is_error', models.BooleanField(default=False)),
                ('error', models.TextField(blank=True, max_length=1000, null=True)),
                ('prev_batch_id', models.CharField(blank=True, max_length=100, null=True)),
                ('batch_id', models.CharField(blank=True, max_length=100, null=True)),
                ('is_consumed', models.BooleanField(default=False)),
                ('is_self', models.BooleanField(default=False)),
                ('site', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, to='sites.Site')),
            ],
            options={
                'ordering': ['timestamp'],
            },
            managers=[
                ('on_site', edc_base.sites.managers.CurrentSiteManager()),
            ],
        ),
        migrations.CreateModel(
            name='OutgoingTransaction',
            fields=[
                ('created', models.DateTimeField(blank=True, default=edc_base.utils.get_utcnow)),
                ('modified', models.DateTimeField(blank=True, default=edc_base.utils.get_utcnow)),
                ('user_created', edc_model_fields.fields.userfield.UserField(blank=True, help_text='Updated by admin.save_model', max_length=50, verbose_name='user created')),
                ('user_modified', edc_model_fields.fields.userfield.UserField(blank=True, help_text='Updated by admin.save_model', max_length=50, verbose_name='user modified')),
                ('hostname_created', models.CharField(blank=True, default=_socket.gethostname, help_text='System field. (modified on create only)', max_length=60)),
                ('hostname_modified', edc_model_fields.fields.hostname_modification_field.HostnameModificationField(blank=True, help_text='System field. (modified on every save)', max_length=50)),
                ('revision', django_revision.revision_field.RevisionField(blank=True, editable=False, help_text='System field. Git repository tag:branch:commit.', max_length=75, null=True, verbose_name='Revision')),
                ('device_created', models.CharField(blank=True, max_length=10)),
                ('device_modified', models.CharField(blank=True, max_length=10)),
                ('id', edc_model_fields.fields.uuid_auto_field.UUIDAutoField(blank=True, editable=False, help_text='System auto field. UUID primary key.', primary_key=True, serialize=False)),
                ('tx', models.BinaryField()),
                ('tx_name', models.CharField(max_length=64)),
                ('tx_pk', models.UUIDField(db_index=True)),
                ('producer', models.CharField(db_index=True, help_text='Producer name', max_length=200)),
                ('action', models.CharField(choices=[('I', 'Insert'), ('U', 'Update'), ('D', 'Delete')], max_length=1)),
                ('timestamp', models.CharField(db_index=True, max_length=50)),
                ('consumed_datetime', models.DateTimeField(blank=True, null=True)),
                ('consumer', models.CharField(blank=True, max_length=200, null=True)),
                ('is_ignored', models.BooleanField(default=False)),
                ('is_error', models.BooleanField(default=False)),
                ('error', models.TextField(blank=True, max_length=1000, null=True)),
                ('prev_batch_id', models.CharField(blank=True, max_length=100, null=True)),
                ('batch_id', models.CharField(blank=True, max_length=100, null=True)),
                ('is_consumed_middleman', models.BooleanField(default=False)),
                ('is_consumed_server', models.BooleanField(default=False)),
                ('using', models.CharField(max_length=25, null=True)),
                ('site', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, to='sites.Site')),
            ],
            options={
                'ordering': ['timestamp'],
            },
            managers=[
                ('on_site', edc_base.sites.managers.CurrentSiteManager()),
            ],
        ),
        migrations.CreateModel(
            name='Server',
            fields=[
                ('created', models.DateTimeField(blank=True, default=edc_base.utils.get_utcnow)),
                ('modified', models.DateTimeField(blank=True, default=edc_base.utils.get_utcnow)),
                ('user_created', edc_model_fields.fields.userfield.UserField(blank=True, help_text='Updated by admin.save_model', max_length=50, verbose_name='user created')),
                ('user_modified', edc_model_fields.fields.userfield.UserField(blank=True, help_text='Updated by admin.save_model', max_length=50, verbose_name='user modified')),
                ('hostname_created', models.CharField(blank=True, default=_socket.gethostname, help_text='System field. (modified on create only)', max_length=60)),
                ('hostname_modified', edc_model_fields.fields.hostname_modification_field.HostnameModificationField(blank=True, help_text='System field. (modified on every save)', max_length=50)),
                ('revision', django_revision.revision_field.RevisionField(blank=True, editable=False, help_text='System field. Git repository tag:branch:commit.', max_length=75, null=True, verbose_name='Revision')),
                ('device_created', models.CharField(blank=True, max_length=10)),
                ('device_modified', models.CharField(blank=True, max_length=10)),
                ('id', edc_model_fields.fields.uuid_auto_field.UUIDAutoField(blank=True, editable=False, help_text='System auto field. UUID primary key.', primary_key=True, serialize=False)),
                ('hostname', models.CharField(max_length=200, unique=True)),
                ('port', models.IntegerField(default='80')),
                ('api_name', models.CharField(default='v1', max_length=15)),
                ('format', models.CharField(default='json', max_length=15)),
                ('authentication', models.CharField(default='api_key', max_length=15)),
                ('is_active', models.BooleanField(default=True)),
                ('last_sync_datetime', models.DateTimeField(blank=True, null=True)),
                ('last_sync_status', models.CharField(blank=True, default='-', max_length=250, null=True)),
                ('comment', models.TextField(blank=True, max_length=50, null=True)),
            ],
            options={
                'ordering': ['hostname', 'port'],
            },
        ),
        migrations.AlterUniqueTogether(
            name='server',
            unique_together={('hostname', 'port')},
        ),
        migrations.AlterUniqueTogether(
            name='history',
            unique_together={('filename', 'hostname')},
        ),
        migrations.AlterUniqueTogether(
            name='client',
            unique_together={('hostname', 'port')},
        ),
    ]



from edc_navbar import Navbar, NavbarItem, site_navbars


navbar = Navbar(name='django_collect_offline')

navbar.append_item(
    NavbarItem(name='synchronization',
               label='Offline Synchronization',
               fa_icon='fa-exchange',
               url_name='django_collect_offline:home_url',
               permission_codename='edc_navbar.nav_synchronization'))

site_navbars.register(navbar)



from edc_base.model_mixins import BaseUuidModel

from .managers import HostManager
from .model_mixins import HostModelMixin


class Server(HostModelMixin, BaseUuidModel):

    """A model to capture the attributes of the server.
    """

    objects = HostManager()

    class Meta:
        ordering = ['hostname', 'port']
        unique_together = (('hostname', 'port'),)



from edc_base.logging import verbose_formatter, file_handler


loggers = {
    'django_collect_offline': {
        'handlers': ['file'],
        'level': 'DEBUG',
        'propagate': True,
    }
}

file_handler['filename'] = f'/tmp/django_collect_offline.log'

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': verbose_formatter,
    },
    'handlers': {
        'file': file_handler
    },
    'loggers': loggers,
}



from django.contrib.sites.models import Site
from django.db import models
from edc_base.model_mixins import BaseUuidModel
from edc_base.sites import CurrentSiteManager, SiteModelMixin
from edc_base.utils import get_utcnow

from .model_mixins import TransactionModelMixin


class OutgoingTransaction(TransactionModelMixin, SiteModelMixin, BaseUuidModel):

    """ Transactions produced locally to be consumed/sent
    to a queue or consumer.
    """

    site = models.ForeignKey(
        Site, on_delete=models.CASCADE, null=True, editable=False)

    is_consumed_middleman = models.BooleanField(
        default=False)

    is_consumed_server = models.BooleanField(
        default=False)

    using = models.CharField(max_length=25, null=True)

    on_site = CurrentSiteManager()

    objects = models.Manager()

    def save(self, *args, **kwargs):
        if not self.using:
            raise ValueError(
                'Value for \'{}.using\' cannot be None.'.format(
                    self._meta.model_name))
        if self.is_consumed_server and not self.consumed_datetime:
            self.consumed_datetime = get_utcnow()
        super().save(*args, **kwargs)

    class Meta:
        ordering = ['timestamp']



import os
import sys

from pathlib import PurePath

from .loggers import LOGGING

APP_NAME = 'django_collect_offline'

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.realpath(__file__))
ETC_DIR = os.path.join(BASE_DIR, 'etc')
SITE_ID = 40
# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.9/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '4w#xs+=lrx4$mmqv+vzy^9i!(sni2eh=q_-9(#w4r20sv($2af'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.sites',
    'django_crypto_fields.apps.AppConfig',
    'django_revision.apps.AppConfig',
    'rest_framework',
    'rest_framework.authtoken',
    'django_js_reverse',
    'simple_history',
    'edc_base.apps.AppConfig',
    'edc_protocol.apps.AppConfig',
    'edc_device.apps.AppConfig',
    'django_collect_offline_files.apps.AppConfig',
    'django_collect_offline.apps.AppConfig',
]


MIDDLEWARE_CLASSES = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    # 'django-crossdomainxhr-middleware.XsSharing',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'simple_history.middleware.HistoryRequestMiddleware',
]

ROOT_URLCONF = 'django_collect_offline.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.media',
            ],
        },
    },
]

WSGI_APPLICATION = 'django_collect_offline.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.9/ref/settings/#databases

DATABASES = {
    # required for tests when acting as a server that deserializes
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    },
    # required for tests when acting as a server but not attempting to
    # deserialize
    'server': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    },
    # required for tests when acting as a client
    'client': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    },
    'test_server': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    },
}

# Password validation
# https://docs.djangoproject.com/en/1.9/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/1.9/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_TZ = True

USE_I18N = True

USE_L10N = True

# USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.9/howto/static-files/

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'static')
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')
MEDIA_URL = '/media/'

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    'django.contrib.staticfiles.finders.DefaultStorageFinder',
)


GIT_DIR = str(PurePath(BASE_DIR).parent)
SHOW_CRYPTO_FORM_DATA = True

LANGUAGES = (
    ('tn', 'Setswana'),
    ('en', 'English'),
)
DEVICE_ID = '15'
SERVER_DEVICE_ID_LIST = ['99']

REST_FRAMEWORK = {
    'PAGE_SIZE': 10,
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.TokenAuthentication',
    ),
}
LOGGING = LOGGING

DJANGO_COLLECT_OFFLINE_SERVER_IP = None
DJANGO_COLLECT_OFFLINE_FILES_REMOTE_HOST = None
DJANGO_COLLECT_OFFLINE_FILES_USER = None
DJANGO_COLLECT_OFFLINE_FILES_USB_VOLUME = None


if 'test' in sys.argv:

    class DisableMigrations:

        def __contains__(self, item):
            return True

        def __getitem__(self, item):
            return None

    MIGRATION_MODULES = DisableMigrations()
    PASSWORD_HASHERS = ('django.contrib.auth.hashers.MD5PasswordHasher', )
    # DEFAULT_FILE_STORAGE = 'inmemorystorage.InMemoryStorage'



import json

from django.apps import apps as django_apps
from django.core.exceptions import MultipleObjectsReturned
from edc_list_data.model_mixins import ListModelMixin
from unittest.case import TestCase

from ..models import OutgoingTransaction
from ..transaction import deserialize


class OfflineTestHelperError(Exception):
    pass


class OfflineTestHelper(TestCase):

    def offline_test_natural_key_attr(self, *app_labels, exclude_models=None):
        """Asserts all models in given apps have a
        natural_key model method.
        """
        exclude_models = exclude_models or []
        for app_label in app_labels:
            models = django_apps.get_app_config(app_label).get_models()
            for model in models:
                if (model._meta.label_lower not in exclude_models
                    and not issubclass(model, ListModelMixin)
                        and '.tests.' not in model.__module__):
                    self.assertTrue(
                        'natural_key' in dir(model),
                        'Model method \'natural_key\' missing. '
                        f'Got \'{model._meta.label_lower}\'.')

    def offline_test_get_by_natural_key_attr(self, *app_labels, exclude_models=None):
        """Asserts all models in given apps have a get_by_natural_key
        manager method.
        """
        exclude_models = exclude_models or []
        for app_label in app_labels:
            models = django_apps.get_app_config(app_label).get_models()
            for model in models:
                if (model._meta.label_lower not in exclude_models
                    and not issubclass(model, ListModelMixin)
                        and '.tests.' not in model.__module__):
                    self.assertTrue(
                        'get_by_natural_key' in dir(model.objects),
                        f'Manager method \'get_by_natural_key\' missing. '
                        f'Got \'{model._meta.label_lower}\'.')

    def offline_test_natural_keys(self, complete_required_crfs):
        """Asserts tuple from natural_key when passed to
        get_by_natural_key successfully gets the model instance.
        """
        for objs in complete_required_crfs.values():
            for obj in objs:
                options = obj.natural_key()
                try:
                    obj.__class__.objects.get_by_natural_key(*options)
                except obj.__class__.DoesNotExist:
                    self.fail(f'get_by_natural_key query failed for '
                              f'\'{obj._meta.label_lower}\' with '
                              f'options {options}.')
                except TypeError as e:
                    raise OfflineTestHelperError(
                        f'{e} See {obj._meta.label_lower}. Got {options}.')

    def offline_test_natural_keys_by_schedule(self, visits=None, visit_attr=None):
        """A wrapper method for offline_test_natural_keys that uses
        the enrollment instance to test each CRF in each visit
        in the schedule linked to the enrollment model.
        """
        complete_required_crfs = {}
        for visit in visits:
            complete_required_crfs.update({
                visit.visit_code: self.complete_required_crfs(
                    visit_code=visit.visit_code,
                    visit=visit,
                    visit_attr=visit_attr,
                    subject_identifier=visit.subject_identifier)})
        self.offline_test_natural_keys(complete_required_crfs)

    def offline_test_serializers_for_visit(self, complete_required_crfs):
        """Assert CRF model instances have transactions and that the
        transactions can be deserialized and compared to their original
        model instances.
        """
        for visit_code in [visit_code for visit_code in complete_required_crfs.keys()]:
            for obj in complete_required_crfs.get(visit_code):
                try:
                    outgoing_transaction = OutgoingTransaction.objects.get(
                        tx_name=obj._meta.label_lower,
                        tx_pk=obj.pk)
                    self.offline_test_deserialize(obj, outgoing_transaction)
                except MultipleObjectsReturned:
                    for outgoing_transaction in OutgoingTransaction.objects.filter(
                            tx_name=obj._meta.label_lower, tx_pk=obj.pk):
                        self.offline_test_deserialize(
                            obj, outgoing_transaction)
                except OutgoingTransaction.DoesNotExist:
                    self.fail('OutgoingTransaction.DoesNotExist unexpectedly '
                              f'raised for {obj._meta.label_lower}')

    def offline_test_deserialize(self, obj, outgoing_transaction):
        """Assert object matches its deserialized transaction.
        """
        json_text = outgoing_transaction.aes_decrypt(outgoing_transaction.tx)
        for deserialised_obj in deserialize(json_text=json_text):
            json_tx = json.loads(json_text)[0]
            self.assertEqual(json_tx.get('model'), obj._meta.label_lower)
            # TODO: verify natural key values?
            self.assertEqual(obj.pk, deserialised_obj.object.pk)



import socket

from django.apps import apps as django_apps
from django.conf import settings
from django.db.models.fields import UUIDField
from django_crypto_fields.constants import LOCAL_MODE
from django_crypto_fields.cryptor import Cryptor
from edc_base.utils import get_utcnow

from .constants import INSERT, UPDATE, DELETE
from .transaction import serialize


class OfflineNaturalKeyMissing(Exception):
    pass


class OfflineGetByNaturalKeyMissing(Exception):
    pass


class OfflineHistoricalManagerError(Exception):
    pass


class OfflineUuidPrimaryKeyMissing(Exception):
    pass


class OfflineModel:

    """A wrapper for offline model instances to add methods called in
    signals for synchronization.
    """

    def __init__(self, instance):
        try:
            self.is_serialized = settings.ALLOW_MODEL_SERIALIZATION
        except AttributeError:
            self.is_serialized = True
        self.instance = instance
        self.has_offline_historical_manager_or_raise()
        self.has_natural_key_or_raise()
        self.has_get_by_natural_key_or_raise()
        self.has_uuid_primary_key_or_raise()

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.instance)})'

    def __str__(self):
        return f'{self.instance._meta.label_lower}'

    def has_natural_key_or_raise(self):
        try:
            self.instance.natural_key
        except AttributeError:
            raise OfflineNaturalKeyMissing(
                f'Model \'{self.instance._meta.app_label}.{self.instance._meta.model_name}\' '
                'is missing method natural_key ')

    def has_get_by_natural_key_or_raise(self):
        try:
            self.instance.__class__.objects.get_by_natural_key
        except AttributeError:
            raise OfflineGetByNaturalKeyMissing(
                f'Model \'{self.instance._meta.app_label}.{self.instance._meta.model_name}\' '
                'is missing manager method get_by_natural_key ')

    def has_offline_historical_manager_or_raise(self):
        """Raises an exception if model uses a history manager and
        historical model history_id is not a UUIDField.

        Note: expected to use edc_base.HistoricalRecords instead of
        simple_history.HistoricalRecords.
        """
        try:
            model = self.instance.__class__.history.model
        except AttributeError:
            model = self.instance.__class__
        field = [
            field for field in model._meta.fields
            if field.name == 'history_id']
        if field and not isinstance(field[0], UUIDField):
            raise OfflineHistoricalManagerError(
                f'Field \'history_id\' of historical model '
                f'\'{model._meta.app_label}.{model._meta.model_name}\' '
                'must be an UUIDfield. '
                'For history = HistoricalRecords() use edc_base.HistoricalRecords instead of '
                'simple_history.HistoricalRecords(). '
                f'See \'{self.instance._meta.app_label}.{self.instance._meta.model_name}\'.')

    def has_uuid_primary_key_or_raise(self):
        if self.primary_key_field.get_internal_type() != 'UUIDField':
            raise OfflineUuidPrimaryKeyMissing(
                f'Expected Model \'{self.instance._meta.label_lower}\' '
                f'primary key {self.primary_key_field} to be a UUIDField '
                f'(e.g. AutoUUIDField). '
                f'Got {self.primary_key_field.get_internal_type()}.')

    @property
    def primary_key_field(self):
        """Return the primary key field.

        Is `id` in most cases. Is `history_id` for Historical models.
        """
        return [field for field in self.instance._meta.fields
                if field.primary_key][0]

    def to_outgoing_transaction(self, using, created=None, deleted=None):
        """ Serialize the model instance to an AES encrypted json object
        and saves the json object to the OutgoingTransaction model.
        """
        OutgoingTransaction = django_apps.get_model(
            'django_collect_offline', 'OutgoingTransaction')
        created = True if created is None else created
        action = INSERT if created else UPDATE
        timestamp_datetime = self.instance.created if created else self.instance.modified
        if not timestamp_datetime:
            timestamp_datetime = get_utcnow()
        if deleted:
            timestamp_datetime = get_utcnow()
            action = DELETE
        outgoing_transaction = None
        if self.is_serialized:
            hostname = socket.gethostname()
            outgoing_transaction = OutgoingTransaction.objects.using(using).create(
                tx_name=self.instance._meta.label_lower,
                tx_pk=getattr(self.instance, self.primary_key_field.name),
                tx=self.encrypted_json(),
                timestamp=timestamp_datetime.strftime('%Y%m%d%H%M%S%f'),
                producer=f'{hostname}-{using}',
                action=action,
                using=using)
        return outgoing_transaction

    def encrypted_json(self):
        """Returns an encrypted json serialized from self.
        """
        json = serialize(objects=[self.instance])
        encrypted_json = Cryptor().aes_encrypt(json, LOCAL_MODE)
        return encrypted_json



import sys

from django.conf import settings

from .client import Client
from .history import History
from .incoming_transaction import IncomingTransaction
from .outgoing_transaction import OutgoingTransaction
from .server import Server

if 'django_collect_offline' in settings.APP_NAME and 'makemigrations' not in sys.argv:
    from ..tests import models



import json
import os
import tempfile

from copy import copy
from django.apps import apps as django_apps
from django.core.serializers.base import DeserializationError
from django.test import TestCase, tag
from edc_base.utils import get_utcnow
from edc_device.constants import NODE_SERVER
from django_collect_offline_files.transaction import TransactionImporter, TransactionExporter
from faker import Faker

from ..models import OutgoingTransaction, IncomingTransaction
from ..site_offline_models import site_offline_models
from ..parsers import datetime_to_date_parser
from ..offline_model import OfflineModel
from ..transaction import TransactionDeserializer, TransactionDeserializerError
from .models import TestModel, TestModelWithFkProtected, TestModelWithM2m, M2m, TestModelDates


fake = Faker()


def my_test_parser(json_text):
    model = 'django_collect_offline.testmodeldates'
    field = 'f2'
    return datetime_to_date_parser(json_text, model=model, field=field)


class TestDeserializer1(TestCase):

    multi_db = True

    def setUp(self):
        site_offline_models.registry = {}
        site_offline_models.loaded = False
        sync_models = [
            'django_collect_offline.testmodel',
            'django_collect_offline.testmodelwithfkprotected',
            'django_collect_offline.testmodelwithm2m',
            'django_collect_offline.testmodelhook', ]
        site_offline_models.register(sync_models)

        self.export_path = os.path.join(tempfile.gettempdir(), 'export')
        if not os.path.exists(self.export_path):
            os.mkdir(self.export_path)
        self.import_path = self.export_path
        IncomingTransaction.objects.all().delete()
        OutgoingTransaction.objects.using('client').all().delete()
        TestModel.objects.all().delete()
        TestModel.objects.using('client').all().delete()
        TestModel.history.all().delete()
        TestModel.history.using('client').all().delete()
        TestModel.objects.using('client').create(f1='model1')
        TestModel.objects.using('client').create(f1='model2')

        tx_exporter = TransactionExporter(
            export_path=self.export_path,
            using='client')
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        self.batch = tx_importer.import_batch(filename=batch.filename)

    def test_deserilize(self):
        tx_deserializer = TransactionDeserializer(override_role=NODE_SERVER)
        tx_deserializer.deserialize_transactions(
            transactions=self.batch.saved_transactions)

    def test_deserilized_object(self):
        tx_deserializer = TransactionDeserializer(override_role=NODE_SERVER)
        tx_deserializer.deserialize_transactions(
            transactions=self.batch.saved_transactions)

    def test_saved(self):
        """Assert transaction object is saved to default db.
        """
        tx_deserializer = TransactionDeserializer(
            override_role=NODE_SERVER,
            using='default')
        tx_deserializer.deserialize_transactions(
            transactions=self.batch.saved_transactions)
        try:
            TestModel.objects.using('default').get(f1='model1')
        except TestModel.DoesNotExist:
            self.fail('TestModel.DoesNotExist unexpectedly raised')


class TestDeserializer2(TestCase):

    multi_db = True

    def setUp(self):
        site_offline_models.registry = {}
        site_offline_models.loaded = False
        sync_models = [
            'django_collect_offline.testmodel',
            'django_collect_offline.testmodelwithfkprotected',
            'django_collect_offline.testmodelwithm2m']
        site_offline_models.register(sync_models, wrapper_cls=OfflineModel)

        self.export_path = os.path.join(tempfile.gettempdir(), 'export')
        if not os.path.exists(self.export_path):
            os.mkdir(self.export_path)
        self.import_path = self.export_path
        M2m.objects.all().delete()
        M2m.objects.using('client').all().delete()
        OutgoingTransaction.objects.using('client').all().delete()
        IncomingTransaction.objects.all().delete()
        TestModel.objects.all().delete()
        TestModel.objects.using('client').all().delete()
        TestModel.history.all().delete()
        TestModel.history.using('client').all().delete()

    def test_saved_on_self(self):
        """Asserts can save on self if allow_self=True.
        """
        TestModel.objects.create(f1='model1')
        tx_exporter = TransactionExporter(export_path=self.export_path)
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        batch = tx_importer.import_batch(filename=batch.filename)
        tx_deserializer = TransactionDeserializer(
            override_role=NODE_SERVER, allow_self=True)
        tx_deserializer.deserialize_transactions(
            transactions=batch.saved_transactions)
        try:
            TestModel.objects.get(f1='model1')
        except TestModel.DoesNotExist:
            self.fail('TestModel unexpectedly does not exist')

    def test_created_from_client(self):
        """Asserts "default" instance is created when "client" instance
        is created.
        """
        TestModel.objects.using('client').create(f1='model1')
        tx_exporter = TransactionExporter(
            export_path=self.export_path, using='client')
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        batch = tx_importer.import_batch(filename=batch.filename)
        tx_deserializer = TransactionDeserializer(
            override_role=NODE_SERVER, allow_self=True)
        tx_deserializer.deserialize_transactions(
            transactions=batch.saved_transactions)
        try:
            TestModel.objects.get(f1='model1')
        except TestModel.DoesNotExist:
            self.fail('TestModel unexpectedly does not exists')

    def test_flagged_as_deserialized(self):
        """Asserts "default" instance is created when "client" instance
        is created.
        """
        TestModel.objects.using('client').create(f1='model1')
        tx_exporter = TransactionExporter(
            export_path=self.export_path, using='client')
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        batch = tx_importer.import_batch(filename=batch.filename)
        tx_deserializer = TransactionDeserializer(
            override_role=NODE_SERVER, allow_self=True)
        tx_deserializer.deserialize_transactions(
            transactions=batch.saved_transactions)
        for transaction in batch.saved_transactions:
            self.assertTrue(transaction.is_consumed)

    def test_deleted_from_client(self):
        """Asserts "default" instance is deleted when "client" instance
        is deleted.
        """
        test_model = TestModel.objects.using('client').create(f1='model1')
        tx_exporter = TransactionExporter(
            export_path=self.export_path, using='client')
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        batch = tx_importer.import_batch(filename=batch.filename)
        tx_deserializer = TransactionDeserializer(
            override_role=NODE_SERVER, allow_self=True)
        tx_deserializer.deserialize_transactions(
            transactions=batch.saved_transactions, deserialize_only=True)
        test_model.delete()
        tx_exporter = TransactionExporter(
            export_path=self.export_path, using='client')
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        batch = tx_importer.import_batch(filename=batch.filename)
        tx_deserializer = TransactionDeserializer(
            override_role=NODE_SERVER, allow_self=True)
        tx_deserializer.deserialize_transactions(
            transactions=batch.saved_transactions)
        try:
            TestModel.objects.get(f1='model1')
        except TestModel.DoesNotExist:
            pass
        else:
            self.fail('TestModel unexpectedly exists')

    def test_dont_allow_saved_on_self(self):
        """Asserts cannot save on self by default.
        """
        TestModel.objects.create(f1='model1')
        tx_exporter = TransactionExporter(export_path=self.export_path)
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        tx_importer.import_batch(filename=batch.filename)
        self.assertRaises(
            TransactionDeserializerError,
            TransactionDeserializer)

    def test_allow_saved_on_self(self):
        """Asserts can save on self by default.
        """
        TestModel.objects.create(f1='model1')
        tx_exporter = TransactionExporter(export_path=self.export_path)
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        tx_importer.import_batch(filename=batch.filename)
        self.assertRaises(TransactionDeserializerError,
                          TransactionDeserializer, allow_self=True)

    def test_allow_saved_on_any_device(self):
        """Asserts can save on self by default.
        """
        TestModel.objects.create(f1='model1')
        tx_exporter = TransactionExporter(export_path=self.export_path)
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        tx_importer.import_batch(filename=batch.filename)
        self.assertRaises(TransactionDeserializerError,
                          TransactionDeserializer, allow_self=True)

    def test_deserialized_with_fk(self):
        """Asserts correctly deserialized model with FK.
        """
        test_model = TestModel.objects.using('client').create(f1='model1')
        TestModelWithFkProtected.objects.using(
            'client').create(f1='f1', test_model=test_model)
        tx_exporter = TransactionExporter(
            export_path=self.export_path, using='client')
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        batch = tx_importer.import_batch(filename=batch.filename)
        tx_deserializer = TransactionDeserializer(
            allow_self=True, override_role=NODE_SERVER)
        tx_deserializer.deserialize_transactions(
            transactions=batch.saved_transactions)
        test_model = TestModel.objects.get(f1='model1')
        try:
            obj = TestModelWithFkProtected.objects.get(f1='f1')
        except TestModelWithFkProtected.DoesNotExist:
            self.fail('TestModel unexpectedly does not exists')
        self.assertEqual(test_model, obj.test_model)

    def test_deserialized_with_history(self):
        """Asserts correctly deserialized model with history.
        """
        TestModel.objects.using('client').create(f1='model1')
        tx_exporter = TransactionExporter(
            export_path=self.export_path, using='client')
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        batch = tx_importer.import_batch(filename=batch.filename)
        tx_deserializer = TransactionDeserializer(
            allow_self=True, override_role=NODE_SERVER)
        tx_deserializer.deserialize_transactions(
            transactions=batch.saved_transactions)
        try:
            TestModel.history.get(f1='model1')
        except TestModel.DoesNotExist:
            self.fail('TestModel history unexpectedly does not exists')

    def test_deserialize_with_m2m(self):
        """Asserts deserializes model with M2M as long as
        M2M instance exists on destination.
        """
        obj = TestModelWithM2m.objects.using('client').create(f1='model1')
        m2m = M2m.objects.using('client').create(name='erik', short_name='bob')
        obj.m2m.add(m2m)

        tx_exporter = TransactionExporter(
            export_path=self.export_path, using='client')
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        batch = tx_importer.import_batch(filename=batch.filename)

        M2m.objects.create(name='erik', short_name='bob')

        tx_deserializer = TransactionDeserializer(
            allow_self=True, override_role=NODE_SERVER)
        tx_deserializer.deserialize_transactions(
            transactions=batch.saved_transactions, deserialize_only=False)
        obj = TestModelWithM2m.objects.get(f1='model1')

        obj.m2m.get(short_name='bob')

    def test_deserialize_with_m2m_missing(self):
        """Asserts deserialization error if m2m instance does not
        exist on destination.
        """
        obj = TestModelWithM2m.objects.using('client').create(f1='model1')
        m2m = M2m.objects.using('client').create(name='erik')
        obj.m2m.add(m2m)
        tx_exporter = TransactionExporter(
            export_path=self.export_path, using='client')
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        batch = tx_importer.import_batch(filename=batch.filename)
        tx_deserializer = TransactionDeserializer(
            allow_self=True, override_role=NODE_SERVER)
        self.assertRaises(
            DeserializationError,
            tx_deserializer.deserialize_transactions,
            transactions=batch.saved_transactions)


class TestDeserializer3(TestCase):

    multi_db = True

    def setUp(self):
        site_offline_models.registry = {}
        site_offline_models.loaded = False
        sync_models = ['django_collect_offline.testmodeldates']
        site_offline_models.register(sync_models, wrapper_cls=OfflineModel)

        self.export_path = os.path.join(tempfile.gettempdir(), 'export')
        if not os.path.exists(self.export_path):
            os.mkdir(self.export_path)
        self.import_path = self.export_path
        IncomingTransaction.objects.all().delete()
        OutgoingTransaction.objects.using('client').all().delete()
        TestModelDates.objects.all().delete()
        TestModelDates.objects.using('client').all().delete()
        self.date = get_utcnow()
        TestModelDates.objects.using('client').create(f2=self.date.date())

        tx_exporter = TransactionExporter(
            export_path=self.export_path,
            using='client')
        batch = tx_exporter.export_batch()
        tx_importer = TransactionImporter(import_path=self.import_path)
        self.batch = tx_importer.import_batch(filename=batch.filename)

        datetime_format = '%Y-%m-%dT%H:%M:%S.%fZ'
        self.obj = IncomingTransaction.objects.all()[0]
        bad_date = self.date.strftime(datetime_format)
        json_text = self.obj.aes_decrypt(self.obj.tx)
        json_obj = json.loads(json_text)
        json_obj[0]['fields']['f2'] = bad_date
        json_text = json.dumps(json_obj)
        self.obj.tx = self.obj.aes_encrypt(json_text)
        self.obj.save()
        self.obj = IncomingTransaction.objects.get(id=self.obj.id)
        self.app_config = copy(django_apps.get_app_config('django_collect_offline'))

    def tearDown(self):
        django_apps.app_configs['django_collect_offline'] = self.app_config

    def test_raises_for_invalid_date(self):
        tx_deserializer = TransactionDeserializer(
            allow_self=True, override_role=NODE_SERVER)
        self.assertRaises(
            DeserializationError,
            tx_deserializer.deserialize_transactions,
            transactions=IncomingTransaction.objects.all())

    def test_custom_parser_declared_in_apps_fixes_date(self):
        app_config = django_apps.get_app_config('django_collect_offline')
        app_config.custom_json_parsers = [my_test_parser]
        django_apps.app_configs['django_collect_offline'] = app_config
        tx_deserializer = TransactionDeserializer(
            allow_self=True, override_role=NODE_SERVER)
        try:
            tx_deserializer.deserialize_transactions(
                transactions=[self.obj])
        except DeserializationError:
            self.fail('DeserializationError unexpectedly raised')



import logging
import os
import tempfile

from django.test.testcases import TestCase
from django_collect_offline_files.action_handler import ActionHandler, ActionHandlerError

logger = logging.getLogger('django_collect_offline')


class TestLogger(TestCase):

    def setUp(self):
        self.src_path = os.path.join(tempfile.gettempdir(), 'src')
        if not os.path.exists(self.src_path):
            os.mkdir(self.src_path)
        self.dst_path = os.path.join(tempfile.gettempdir(), 'dst')
        if not os.path.exists(self.dst_path):
            os.mkdir(self.dst_path)
        self.archive_path = os.path.join(tempfile.gettempdir(), 'archive')
        if not os.path.exists(self.archive_path):
            os.mkdir(self.archive_path)

    def test_action_logging(self):
        """Asserts an invalid action is logged.
        """
        kwargs = dict(
            using='client',
            src_path=self.src_path,
            dst_path=self.dst_path,
            archive_path=self.archive_path,
            remote_host='localhost')

        action = 'BLAH'

        action_handler = ActionHandler(**kwargs)

        with self.assertLogs(logger, logging.ERROR) as cm:
            try:
                action_handler.action(label=action)
            except ActionHandlerError as e:
                logger.warn(e)
                logger.exception(e)
        self.assertIn('BLAH', ' '.join(cm.output))



from django.db import models
from django.db.models.deletion import PROTECT
from edc_base.model_managers import HistoricalRecords
from edc_base.model_mixins import BaseUuidModel, BaseModel
from edc_base.utils import get_utcnow
from edc_list_data.model_mixins import ListModelMixin
from simple_history.models import HistoricalRecords as BadHistoricalRecords
from uuid import uuid4


class TestModelManager(models.Manager):

    def get_by_natural_key(self, f1):
        return self.get(f1=f1)


class TestModel(BaseUuidModel):

    f1 = models.CharField(max_length=10, unique=True)

    f2 = models.CharField(max_length=10, null=True)

    f3 = models.CharField(max_length=10, default=uuid4())

    objects = TestModelManager()

    history = HistoricalRecords()

    def natural_key(self):
        return (self.f1,)


class TestModelDates(BaseUuidModel):

    f1 = models.CharField(max_length=10, unique=True)

    f2 = models.DateField(
        blank=True,
        null=True)

    f3 = models.DateTimeField(
        blank=True,
        default=get_utcnow)

    objects = TestModelManager()

    def natural_key(self):
        return (self.f1,)


class BadTestModel(BaseUuidModel):
    """A test model that is missing natural_key and get_by_natural_key.
    """

    f1 = models.CharField(max_length=10, default='f1')

    objects = models.Manager()


class AnotherBadTestModel(BaseUuidModel):
    """A test model that is missing get_by_natural_key.
    """

    f1 = models.CharField(max_length=10, default='f1')

    objects = models.Manager()

    def natural_key(self):
        return (self.f1,)


class YetAnotherBadTestModel(BaseUuidModel):
    """A test model with the wrong HistoricalManager.
    """

    f1 = models.CharField(max_length=10, default='f1')

    objects = TestModelManager()

    history = BadHistoricalRecords()  # missing UUID handling

    def natural_key(self):
        return (self.f1,)


class TestOfflineModelNoHistoryManager(BaseUuidModel):
    """A test model without a HistoricalManager.
    """

    f1 = models.CharField(max_length=10, default='f1')

    objects = TestModelManager()

    def natural_key(self):
        return (self.f1,)


class TestOfflineModelNoUuid(BaseModel):
    """A test model without a HistoricalManager.
    """

    f1 = models.CharField(max_length=10, default='f1')

    objects = TestModelManager()

    def natural_key(self):
        return (self.f1,)


class M2m(ListModelMixin, BaseUuidModel):

    pass


class TestModelWithFkProtected(BaseUuidModel):

    f1 = models.CharField(max_length=10, unique=True)

    test_model = models.ForeignKey(TestModel, on_delete=PROTECT)

    objects = TestModelManager()

    history = HistoricalRecords()

    def natural_key(self):
        return (self.f1,)
    natural_key.dependencies = ['django_collect_offline.test_model']


class TestModelWithM2m(BaseUuidModel):

    f1 = models.CharField(max_length=10, unique=True)

    m2m = models.ManyToManyField(M2m)

    objects = TestModelManager()

    history = HistoricalRecords()

    def natural_key(self):
        return (self.f1,)



from django.core import serializers


def serialize(objects=None):
    """A simple wrapper of Django's serializer with defaults
    for JSON and natural keys.

    Note: use_natural_primary_keys is False as once
    a pk is set, it should not be changed throughout the
    distributed data.
    """
    return serializers.serialize(
        'json', objects,
        ensure_ascii=True,
        use_natural_foreign_keys=True,
        use_natural_primary_keys=False)



from django.contrib.sites.models import Site
from django.db import models
from edc_base.model_mixins import BaseUuidModel
from edc_base.sites import CurrentSiteManager, SiteModelMixin

from .model_mixins import TransactionModelMixin


class IncomingTransaction(TransactionModelMixin, SiteModelMixin, BaseUuidModel):

    """ Transactions received from a remote host.
    """

    site = models.ForeignKey(
        Site, on_delete=models.CASCADE, null=True, editable=False)

    is_consumed = models.BooleanField(
        default=False)

    is_self = models.BooleanField(
        default=False)

    on_site = CurrentSiteManager()

    objects = models.Manager()

    class Meta:
        ordering = ['timestamp']



from django.contrib.auth.models import User
from django.test import TestCase, tag
from edc_base.site_models import SiteModelAlreadyRegistered, SiteModelNotRegistered

from ..site_offline_models import site_offline_models
from .models import TestModel


class TestSiteSyncModels(TestCase):

    def setUp(self):
        site_offline_models.registry = {}
        site_offline_models.loaded = False
        self.offline_models = ['django_collect_offline.testmodel',
                               'django_collect_offline.badtestmodel',
                               'django_collect_offline.anotherbadtestmodel',
                               'django_collect_offline.yetanotherbadtestmodel',
                               'django_collect_offline.testmodelwithfkprotected',
                               'django_collect_offline.testmodelwithm2m',
                               'django_collect_offline.testofflinemodelnohistorymanager',
                               'django_collect_offline.testofflinemodelnouuid']
        site_offline_models.register(models=self.offline_models)

    def test_site_offline_models(self):
        self.assertTrue(str(site_offline_models))
        self.assertTrue(repr(site_offline_models))

    def test_site_offline_models2(self):
        self.assertIn('django_collect_offline', site_offline_models.site_models())

    def test_already_registered(self):
        self.assertRaises(
            SiteModelAlreadyRegistered,
            site_offline_models.register,
            models=['django_collect_offline.testmodel'])

    def test_get_as_sync_model_ok(self):
        self.assertTrue(
            site_offline_models.get_wrapped_instance(TestModel()))

    def test_get_as_sync_model_not_registered(self):
        user = User()
        self.assertRaises(
            SiteModelNotRegistered,
            site_offline_models.get_wrapped_instance, user)

    def test_get_as_sync_model_none(self):
        self.assertRaises(
            AttributeError,
            site_offline_models.get_wrapped_instance)



import copy
import sys

from django.apps import apps as django_apps
from django.utils.module_loading import import_module, module_has_submodule


class SiteModelAlreadyRegistered(Exception):
    pass


class SiteModelError(Exception):
    pass


class SiteModelNotRegistered(Exception):
    pass


class SiteSyncModelError(Exception):
    pass


class SiteModels:

    wrapper_cls = None
    register_historical = True

    def __init__(self):
        self.registry = {}
        self.loaded = False

    def register(self, models=None, wrapper_cls=None):
        """Registers with app_label.modelname, wrapper_cls.
        """
        self.loaded = True
        for model in models:
            model = model.lower()
            if model not in self.registry:
                self.registry.update({model: wrapper_cls or self.wrapper_cls})
                if self.register_historical:
                    historical_model = '.historical'.join(model.split('.'))
                    self.registry.update(
                        {historical_model: wrapper_cls or self.wrapper_cls})
            else:
                raise SiteModelAlreadyRegistered(
                    f'Model is already registered. Got {model}.')

    def register_for_app(self, app_label=None, exclude_models=None,
                         include_list_models=None):
        """Registers all models for this app_label.
        """
        from edc_list_data.model_mixins import ListModelMixin
        models = []
        exclude_models = exclude_models or []
        app_config = django_apps.get_app_config(app_label)
        for model in app_config.get_models():
            if model._meta.label_lower in exclude_models:
                pass
            elif include_list_models and issubclass(model, ListModelMixin):
                models.append(model._meta.label_lower)
            else:
                models.append(model._meta.label_lower)
        self.register(models)

    def get_wrapped_instance(self, instance=None):
        """Returns a wrapped model instance.
        """
        if instance._meta.label_lower not in self.registry:
            raise SiteModelNotRegistered(
                f'{repr(instance)} is not registered with {self}.')
        wrapper_cls = self.registry.get(
            instance._meta.label_lower) or self.wrapper_cls
        if wrapper_cls:
            return wrapper_cls(instance)
        return instance

    def site_models(self, app_label=None):
        """Returns a dictionary of registered models.
        """
        site_models = {}
        app_configs = (django_apps.get_app_configs()
                       if app_label is None else [django_apps.get_app_config(app_label)])
        for app_config in app_configs:
            model_list = [model for model in app_config.get_models()
                          if model._meta.label_lower in self.registry]
            if model_list:
                model_list.sort(key=lambda m: m._meta.verbose_name)
                site_models.update({app_config.name: model_list})
        return site_models

    def autodiscover(self, module_name=None):
        module_name = module_name or self.module_name
        sys.stdout.write(' * checking for models to register ...\n')
        for app in django_apps.app_configs:
            try:
                mod = import_module(app)
                try:
                    before_import_registry = copy.deepcopy(self.registry)
                    import_module(f'{app}.{module_name}')
                    sys.stdout.write(
                        f' * registered models from \'{app}\'.\n')
                except Exception as e:
                    if f'No module named \'{app}.{module_name}\'' not in str(e):
                        raise
                    self.registry = before_import_registry
                    if module_has_submodule(mod, module_name):
                        raise
            except ImportError:
                pass


class SiteOfflineModels(SiteModels):

    """ Main controller of :class:`offline_models` objects.
    """

    module_name = 'offline_models'
    register_historical = True

    @property
    def wrapper_cls(self):
        from .offline_model import OfflineModel
        return OfflineModel


site_offline_models = SiteOfflineModels()



import json

from django.apps import apps as django_apps
from django.contrib.auth.decorators import login_required
from django.core.serializers.json import Serializer
from django.utils.decorators import method_decorator
from django.views.generic.base import TemplateView
from django_crypto_fields.constants import LOCAL_MODE
from django_crypto_fields.cryptor import Cryptor
from edc_base.view_mixins import EdcBaseViewMixin


class RenderView(EdcBaseViewMixin, TemplateView):

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def get_template_names(self):
        return 'django_collect_offline/render_{self.kwargs.get("model_name")}.html'

    @property
    def model(self):
        model_name = self.kwargs.get('model_name')
        return django_apps.get_model('django_collect_offline', model_name)

    @property
    def queryset(self):
        pk = self.kwargs.get('pk')
        return self.model.objects.filter(pk=pk)

    @property
    def json_tx(self):
        cryptor = Cryptor()
        return json.loads(cryptor.aes_decrypt(
            self.queryset.first().tx, mode=LOCAL_MODE))

    @property
    def json_obj(self):
        serializer = Serializer()
        return json.loads(serializer.serialize(
            self.queryset, use_natural_primary_keys=False))

    def get_context_data(self, **kwargs):
        context = super(RenderView, self).get_context_data(**kwargs)
        context.update(json_tx=self.json_tx[0])
        context.update(json_obj=self.json_obj[0])
        return context



import requests

from datetime import datetime
from django.apps import apps as django_apps
from django.urls import reverse
from django.views.generic.base import TemplateView
from edc_base.view_mixins import EdcBaseViewMixin
from django_collect_offline_files.models import ImportedTransactionFileHistory
from requests.exceptions import ConnectionError, HTTPError

from ..admin import django_collect_offline_admin
from ..offline_view_mixin import OfflineViewMixin
from ..models import Client


class SyncReportClientView(
        EdcBaseViewMixin, OfflineViewMixin, TemplateView):

    template_name = 'django_collect_offline/sync_report_client.html'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        app_config = django_apps.get_app_config('edc_map')  # FIXME: ?
        context.update(
            django_collect_offline_admin=django_collect_offline_admin,
            project_name=context.get(
                'project_name') + ': ' + self.role.title(),
            base_template_name=app_config.base_template_name)
        return context

    def get(self, request, *args, **kwargs):
        report = Report()
        context = self.get_context_data(**kwargs)
        context.update({'report_data': report.report_data})
        if request.is_ajax():
            action = request.GET.get('action')
            # response_data = {}
            if action == 'receive':
                # TODO: see TransactionExporterViewMixin
                raise TypeError()
#                 response_data = dict(
#                     confirm=True, hostname=sync_confirm.confirm_code)
#                 return HttpResponse(
# json.dumps(response_data), content_type='application/')
        return self.render_to_response(context)


class Report:
    """ Displays number of pending transactions in the client and number of
    times the machine have synced.
    """
    imported_history_model = ImportedTransactionFileHistory

    def __init__(self):
        self.report_data = []
        for client in Client.objects.all():
            try:
                self.history_model.objects.get(
                    hostname=client.hostname,
                    confirmed_date=datetime.today().date())  # ????
                received = True
            except self.history_model.DoesNotExist:
                received = False
            try:
                url = 'http://' + client.hostname + reverse(
                    'django_collect_offline:transaction-count')
                r = requests.get(url, timeout=3)
                data = r.json()
                connected = True
            except (ConnectionError, Exception):
                pending = -1
                connected = False
            except HTTPError:
                pending = -1
                connected = True
            else:
                pending = data.get('outgoingtransaction_count')
                pending_files_no = data.get('pending_files_no')
            data = {'device': client.hostname,
                    'sync_times': self.synced_files(client.hostname),
                    'pending': pending,
                    'connected': connected,
                    'received': received,
                    'pending_files_no': pending_files_no,
                    'comment': client.comment}
            self.report_data.append(data)

    def synced_files(self, hostname):
        producer = f'{hostname}-default'
        return [p for p in self.imported_history_model.objects.filter(
            producer=producer,
            created__date=datetime.today().date()).order_by('-created')]



import socket

from django.apps import apps as django_apps
from django_crypto_fields.constants import LOCAL_MODE
from django_crypto_fields.cryptor import Cryptor
from django_collect_offline_files.transaction.file_archiver import FileArchiver
from edc_device.constants import NODE_SERVER, CENTRAL_SERVER

from ..constants import DELETE
from ..models import IncomingTransaction
from .deserialize import deserialize


class TransactionDeserializerError(Exception):
    pass


def save(obj=None, m2m_data=None):
    """Saves a deserialized model object.

    Uses save_base to avoid running code in model.save() and
    to avoid triggering signals (if raw=True).
    """
    m2m_data = {} if m2m_data is None else m2m_data
    obj.save_base(raw=True)
    for attr, values in m2m_data.items():
        for value in values:
            getattr(obj, attr).add(value)


def aes_decrypt(cipher_text):
    return Cryptor().aes_decrypt(cipher_text, LOCAL_MODE)


class TransactionDeserializer:

    def __init__(self, using=None, allow_self=None, override_role=None, **kwargs):
        app_config = django_apps.get_app_config('edc_device')
        self.aes_decrypt = aes_decrypt
        self.deserialize = deserialize
        self.save = save
        self.allow_self = allow_self
        self.using = using
        if not app_config.is_server:
            if override_role not in [NODE_SERVER, CENTRAL_SERVER]:
                raise TransactionDeserializerError(
                    'Transactions may only be deserialized on a server. '
                    f'Got override_role={override_role}, device={app_config.device_id}, '
                    f'device_role={app_config.device_role}.')

    def deserialize_transactions(self, transactions=None, deserialize_only=None):
        """Deserializes the encrypted serialized model
        instances, tx, in a queryset of transactions.

        Note: each transaction instance contains encrypted JSON text
        that represents just ONE model instance.
        """

        if not self.allow_self and transactions.filter(
                producer=socket.gethostname()).exists():
            raise TransactionDeserializerError(
                f'Not deserializing own transactions. Got '
                f'allow_self=False, hostname={socket.gethostname()}')
        for transaction in transactions:
            json_text = self.aes_decrypt(cipher_text=transaction.tx)
            json_text = self.custom_parser(json_text)
            deserialized = next(self.deserialize(json_text=json_text))
            if not deserialize_only:
                if transaction.action == DELETE:
                    deserialized.object.delete()
                else:
                    self.save(
                        obj=deserialized.object,
                        m2m_data=deserialized.m2m_data)
                transaction.is_consumed = True
                transaction.save()

    def custom_parser(self, json_text=None):
        """Runs json_text thru custom parsers.
        """
        app_config = django_apps.get_app_config('django_collect_offline')
        for json_parser in app_config.custom_json_parsers:
            json_text = json_parser(json_text)
        return json_text


class CustomTransactionDeserializer(TransactionDeserializer):

    file_archiver_cls = FileArchiver

    def __init__(self,
                 using=None, allow_self=None, override_role=None,
                 order_by=None, model=None, batch=None, producer=None,
                 **options):
        super().__init__(**options)
        self.allow_self = allow_self
        self.aes_decrypt = aes_decrypt
        self.deserialize = deserialize
        self.override_role = override_role
        self.save = save
        self.using = using
        """ Find how inherit parent properties.
        """
        filters = {}
        if model:
            filters.update({'tx_name': model})
        if batch:
            filters.update({'batch_id': batch})
        if producer:
            filters.update({'producer': producer})
        if filters:
            try:
                transactions = IncomingTransaction.objects.filter(
                    **filters).order_by(*order_by.split(','))
                self.deserialize_transactions(transactions=transactions)
            except TransactionDeserializerError as e:
                raise TransactionDeserializerError(e) from e
            else:
                obj = self.file_archiver_cls(
                    src_path=django_apps.get_app_config(
                        'django_collect_offline').pending_folder,
                    dst_path=django_apps.get_app_config(
                        'django_collect_offline').archive_folder)
                obj.archive(filename=f'{batch}.json')



from django.conf import settings
from django.urls import path, re_path, include
from edc_constants.constants import UUID_PATTERN
from rest_framework.authtoken.views import obtain_auth_token
from rest_framework.routers import DefaultRouter

from .admin import django_collect_offline_admin
from .views import HomeView, RenderView
from .views import OutgoingTransactionViewSet, IncomingTransactionViewSet
from .views import TransactionCountView, OfflineReportView


router = DefaultRouter()
router.register(r'outgoingtransaction', OutgoingTransactionViewSet)
router.register(r'incomingtransaction', IncomingTransactionViewSet)

app_name = 'django_collect_offline'

urlpatterns = [
    path('admin/', django_collect_offline_admin.urls),
    path('api/transaction-count/',
         TransactionCountView.as_view(), name='transaction-count'),
    path('sync-report/',
         OfflineReportView.as_view(), name='sync-report'),
    path('api/', include(router.urls)),
    path('api-token-auth/', obtain_auth_token),
    re_path(f'render/(?P<model_name>\w+)/(?P<pk>{UUID_PATTERN.pattern})/',
            RenderView.as_view(), name='render_url'),
    path('', HomeView.as_view(), name='home_url'),
]



from .deserialize import deserialize
from .serialize import serialize
from .transaction_deserializer import (
    TransactionDeserializer, TransactionDeserializerError,
    CustomTransactionDeserializer)



from django.db import models
from edc_base.model_mixins import BaseUuidModel
from edc_base.utils import get_utcnow

from .managers import HistoryManager


# FIXME: is this model used?
class History(BaseUuidModel):

    objects = HistoryManager()

    filename = models.CharField(
        max_length=100,
        unique=True)

    hostname = models.CharField(
        max_length=100)

    sent_datetime = models.DateTimeField(default=get_utcnow)

    def __str__(self):
        return f'</{self.filename}.{self.hostname}>'

    def natural_key(self):
        return (self.filename, self.hostname)

    class Meta:
        ordering = ('-sent_datetime',)
        verbose_name = 'Sent History'
        verbose_name_plural = 'Sent History'
        unique_together = (('filename', 'hostname'),)



from edc_rest.binary_field import BinaryField
from edc_rest.serializers import BaseModelSerializerMixin
from rest_framework import serializers

from .choices import ACTIONS
from .models import IncomingTransaction, OutgoingTransaction


class BaseTransactionSerializer(BaseModelSerializerMixin, serializers.Serializer):

    pk = serializers.UUIDField(read_only=True)

    tx = BinaryField()

    tx_name = serializers.CharField(
        max_length=64)

    tx_pk = serializers.UUIDField()

    producer = serializers.CharField(max_length=200)

    action = serializers.ChoiceField(
        choices=ACTIONS)

    timestamp = serializers.CharField(
        max_length=50)

    consumed_datetime = serializers.DateTimeField(
        allow_null=True,
        default=None)

    consumer = serializers.CharField(
        max_length=200,
        allow_null=True,
        allow_blank=True)

    is_ignored = serializers.BooleanField(
        default=False)

    is_error = serializers.BooleanField(
        default=False)

    error = serializers.CharField(
        max_length=1000,
        allow_null=True,
        allow_blank=True)


class IncomingTransactionSerializer(BaseTransactionSerializer):

    model_class = IncomingTransaction

    is_consumed = serializers.BooleanField(
        default=False)


class OutgoingTransactionSerializer(BaseTransactionSerializer):

    model_class = OutgoingTransaction

    is_consumed_server = serializers.BooleanField(
        default=False)

    is_consumed_middleman = serializers.BooleanField(
        default=False)



from .home_view import HomeView
from .render_view import RenderView
from .offline_report_view import OfflineReportView
from .transaction_count_view import TransactionCountView
from .view_sets import (
    OutgoingTransactionViewSet, IncomingTransactionViewSet)



from django.test import TestCase, tag
from django.utils.six import BytesIO
from django_crypto_fields.constants import LOCAL_MODE
from django_crypto_fields.cryptor import Cryptor
from rest_framework.parsers import JSONParser
from rest_framework.renderers import JSONRenderer

from ..models import OutgoingTransaction
from ..serializers import OutgoingTransactionSerializer
from ..site_offline_models import site_offline_models
from .models import TestModel


class TestSerializers(TestCase):

    multi_db = True

    def setUp(self):
        TestModel.objects.create(f1='give any one species too much rope ...')
        site_offline_models.registry = {}
        site_offline_models.loaded = False
        sync_models = ['django_collect_offline.testmodel']
        site_offline_models.register(sync_models)

    def test_outgoingtransaction_serializer_inits(self):
        obj = OutgoingTransaction.objects.last()
        serializer = OutgoingTransactionSerializer(obj)
        serializer.data

    def test_outgoingtransaction_serializer_renders(self):
        obj = OutgoingTransaction.objects.last()
        serializer = OutgoingTransactionSerializer(obj)
        serializer.data
        self.assertTrue(JSONRenderer().render(serializer.data))

    def test_outgoingtransaction_serializer_parses(self):
        obj = OutgoingTransaction.objects.last()
        serializer = OutgoingTransactionSerializer(obj)
        serializer.data
        content = JSONRenderer().render(serializer.data)
        stream = BytesIO(content)
        data = JSONParser().parse(stream)
        self.assertEqual(data['tx'], serializer.data['tx'])

    def test_outgoingtransaction_serializer_validates(self):
        obj = OutgoingTransaction.objects.last()
        serializer = OutgoingTransactionSerializer(obj)
        serializer.data
        content = JSONRenderer().render(serializer.data)
        stream = BytesIO(content)
        data = JSONParser().parse(stream)
        serializer = OutgoingTransactionSerializer(data=data)
        serializer.is_valid()
        self.assertEqual(obj.tx, serializer.validated_data['tx'])

    def test_outgoingtransaction_serializer_tx_decrypts(self):
        obj = OutgoingTransaction.objects.last()
        serializer = OutgoingTransactionSerializer(obj)
        serializer.data
        content = JSONRenderer().render(serializer.data)
        stream = BytesIO(content)
        data = JSONParser().parse(stream)
        serializer = OutgoingTransactionSerializer(data=data)
        serializer.is_valid()
        cryptor = Cryptor()
        self.assertTrue(cryptor.aes_decrypt(
            serializer.validated_data['tx'], LOCAL_MODE))
        value = cryptor.aes_decrypt(
            serializer.validated_data['tx'], LOCAL_MODE).encode()
        stream = BytesIO(value)
        json_data = JSONParser().parse(stream)
        self.assertTrue(json_data[0]['fields']['f1'],
                        'give any one species too much rope ...')



from django.contrib import admin

from rest_framework.authtoken.models import Token
from rest_framework.authtoken.admin import TokenAdmin

from .admin_site import django_collect_offline_admin
from .models import IncomingTransaction, OutgoingTransaction, Client, Server

# registering TokenAdmin with model Token fails
# if you have not declared 'rest_framework.authtoken' in INSTALLED_APPS.

# unregister to re-register with our admin class
admin.site.unregister(Token)


@admin.register(Token, site=django_collect_offline_admin)
class MyTokenAdmin(TokenAdmin):
    raw_id_fields = ('user',)
    list_display = ('key', 'user', 'created')
    fields = ('user', )
    ordering = ('-created',)


@admin.register(IncomingTransaction, site=django_collect_offline_admin)
class IncomingTransactionAdmin (admin.ModelAdmin):

    ordering = ('-timestamp', )

    list_display = (
        'tx_name', 'view', 'producer', 'is_consumed', 'is_error',
        'is_ignored', 'consumer', 'consumed_datetime', 'action',
        'tx_pk', 'timestamp', 'hostname_modified')

    list_filter = (
        'is_consumed', 'is_error', 'is_ignored', 'consumer',
        'consumed_datetime', 'producer', 'action', 'tx_name',
        'hostname_modified')

    search_fields = ('tx_pk', 'tx', 'timestamp', 'error', 'id')


@admin.register(OutgoingTransaction, site=django_collect_offline_admin)
class OutgoingTransactionAdmin (admin.ModelAdmin):

    ordering = ('-timestamp', )

    list_display = (
        'tx_name', 'view', 'producer',
        'is_consumed_server', 'is_error', 'consumer',
        'consumed_datetime', 'action', 'tx_pk', 'timestamp',
        'hostname_modified')

    list_filter = (
        'is_error', 'is_consumed_server',
        'consumer', 'consumed_datetime', 'producer',
        'action', 'tx_name', 'hostname_modified')

    search_fields = ('tx_pk', 'tx', 'timestamp', 'error', 'id')


class HostAdmin(admin.ModelAdmin):

    list_display = (
        'hostname', 'port', 'is_active',
        'last_sync_datetime', 'last_sync_status', 'comment')

    list_filter = ('is_active', 'last_sync_datetime', 'last_sync_status',)


@admin.register(Client, site=django_collect_offline_admin)
class ClientAdmin(HostAdmin):
    pass


@admin.register(Server, site=django_collect_offline_admin)
class ServerAdmin(HostAdmin):
    pass



from django.contrib.auth.models import User

from rest_framework_httpsignature.authentication import SignatureAuthentication


class OfflineSignatureAuthentication(SignatureAuthentication):

    # The HTTP header used to pass the consumer key ID.
    API_KEY_HEADER = 'X-Api-Key'

    # A method to fetch (User instance, user_secret_string) from the
    # consumer key ID, or None in case it is not found.
    def fetch_user_data(self, username, api_key):
        try:
            user = User.objects.get(username=username, api_key=api_key)
            return (user, user.secret)
        except User.DoesNotExist:
            return None



import json

from django.core.serializers.json import DjangoJSONEncoder


def datetime_to_date_parser(json_text, model=None, field=None):

    if model and field:
        try:
            json_data = json.loads(json_text)
        except TypeError:
            pass
        else:
            fields = json_data[0]['fields']
            if model == json_data[0]['model'] and field in fields:
                json_data[0]['fields'][field] = fields[field][:10]
                json_text = json.dumps(json_data, cls=DjangoJSONEncoder)
    return json_text



from django.db.models.signals import post_save, m2m_changed, post_delete
from django.dispatch import receiver
from edc_base.site_models import SiteModelNotRegistered
from rest_framework.authtoken.models import Token

from .site_offline_models import site_offline_models


@receiver(post_save, sender=Token)
def create_auth_token(sender, instance, raw, created, **kwargs):
    """Create token when a user is created (from rest_framework).
    """
    if not raw:
        if created:
            sender.objects.create(user=instance)


@receiver(m2m_changed, weak=False, dispatch_uid='serialize_m2m_on_save')
def serialize_m2m_on_save(sender, action, instance, using, **kwargs):
    """ Part of the serialize transaction process that ensures m2m are
    serialized correctly.

    Skip those not registered.
    """
    if action == 'post_add':
        try:
            wrapped_instance = site_offline_models.get_wrapped_instance(
                instance)
        except SiteModelNotRegistered:
            pass
        else:
            wrapped_instance.to_outgoing_transaction(using, created=True)


@receiver(post_save, weak=False, dispatch_uid='serialize_on_save')
def serialize_on_save(sender, instance, raw, created, using, **kwargs):
    """ Serialize the model instance as an OutgoingTransaction.

    Skip those not registered.
    """
    if not raw:
        try:
            wrapped_instance = site_offline_models.get_wrapped_instance(
                instance)
        except SiteModelNotRegistered:
            pass
        else:
            wrapped_instance.to_outgoing_transaction(using, created=created)


@receiver(post_delete, weak=False, dispatch_uid="serialize_on_post_delete")
def serialize_on_post_delete(sender, instance, using, **kwargs):
    """Creates a serialized OutgoingTransaction when
    a model instance is deleted.

    Skip those not registered.
    """
    try:
        wrapped_instance = site_offline_models.get_wrapped_instance(instance)
    except SiteModelNotRegistered:
        pass
    else:
        wrapped_instance.to_outgoing_transaction(
            using, created=False, deleted=True)



from edc_base.model_mixins import BaseUuidModel

from .model_mixins import HostModelMixin
from .managers import HostManager


class Client(HostModelMixin, BaseUuidModel):

    """A model to capture the attributes of hosts (clients) to be
    contacted by the server.
    """

    objects = HostManager()

    class Meta:
        ordering = ['hostname', 'port']
        unique_together = (('hostname', 'port'),)



from edc_base.view_mixins import EdcBaseViewMixin

from django.apps import apps as django_apps
from django.db.models.aggregates import Count
from django.views.generic import ListView

from django_collect_offline_files.admin_site import django_collect_offline_files_admin

from ..offline_view_mixin import OfflineViewMixin
from ..models import IncomingTransaction


class OfflineReportView(EdcBaseViewMixin, OfflineViewMixin, ListView):

    template_name = 'django_collect_offline/offline_report.html'

    def get_queryset(self):
        return IncomingTransaction.objects.values('producer').filter(
            is_consumed=False).annotate(
                pending=Count('is_consumed')).order_by('producer')

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        app_config = django_apps.get_app_config('django_collect_offline')
        context.update(
            base_template_name=app_config.base_template_name,
            django_collect_offline_files_admin=django_collect_offline_files_admin)
        return context



from django.core import serializers
from django.test import TestCase, tag

from ..transaction import deserialize
from ..parsers import datetime_to_date_parser
from .models import TestModelDates


class TestParsers(TestCase):

    def test_date_parser(self):
        datetime_format = '%Y-%m-%dT%H:%M:%S.%fZ'
        obj = TestModelDates.objects.create()
        json_text = serializers.serialize('json', [obj])
        bad_date = obj.f3.strftime(datetime_format)
        json_text = json_text.replace('null', f'"{bad_date}"')
        json_text = datetime_to_date_parser(
            json_text, model='django_collect_offline.testmodeldates', field='f2')
        json_data = deserialize(json_text)
        for obj in json_data:
            self.assertEqual(obj.object.f3.date(), obj.object.f2)

    def test_date_parser_with_null(self):
        obj = TestModelDates.objects.create()
        json_text = serializers.serialize('json', [obj])
        json_text = datetime_to_date_parser(json_text, field='f2')
        json_data = deserialize(json_text)
        for obj in json_data:
            self.assertIsNone(obj.object.f2)

    def test_date_parser_with_none(self):
        try:
            datetime_to_date_parser(
                None, model='django_collect_offline.testmodeldates', field='f2')
        except TypeError:
            self.fail(
                'the JSON object must be str, bytes or bytearry, not NoneType')

    def test_date_parser_with_bad_fieldname(self):
        obj = TestModelDates.objects.create()
        json_text = serializers.serialize('json', [obj])
        self.assertEqual(json_text, datetime_to_date_parser(
            json_text, field='blah'))

    def test_date_parser_with_no_fieldname(self):
        obj = TestModelDates.objects.create()
        json_text = serializers.serialize('json', [obj])
        self.assertEqual(json_text, datetime_to_date_parser(json_text))



from django.db import models


class HostManager(models.Manager):

    def get_by_natural_key(self, hostname, port):
        return self.get(hostname=hostname, port=port)


class HistoryManager(models.Manager):

    def get_by_natural_key(self, filename, sent_datetime):
        return self.get(filename=filename, sent_datetime=sent_datetime)



import json

from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist
from edc_device.constants import CENTRAL_SERVER, NODE_SERVER
from json.decoder import JSONDecodeError
from requests.exceptions import RequestException
from rest_framework.authtoken.models import Token

from .constants import SERVER, CLIENT
from .models import Client, Server


class OfflineViewMixin:

    @property
    def device_role(self):
        edc_device_app = django_apps.get_app_config('edc_device')
        return edc_device_app.device_role

    @property
    def host_model(self):
        host_model = None
        if self.device_role in [SERVER, CENTRAL_SERVER]:
            host_model = Server
        elif self.device_role in [SERVER, NODE_SERVER]:
            host_model = Client
        elif self.device_role == CLIENT:
            host_model = Server
        else:
            raise ImproperlyConfigured(
                'Project uses \'django_collect_offline\' but has not defined a '
                'valid device role for this '
                f'app instance. See AppConfig. Got {self.device_role}.')
        return host_model

    @property
    def resource(self):
        resource = 'outgoingtransaction'
        if self.device_role in [SERVER, CENTRAL_SERVER, NODE_SERVER]:
            resource = 'outgoingtransaction'
        elif self.device_role == CLIENT:
            resource = 'incomingtransaction'
        else:
            raise ImproperlyConfigured(
                'Project uses \'django_collect_offline\' but has '
                f'not defined a valid role for this '
                f'app instance. See AppConfig. Got {self.device_role}.')
        return resource

    @property
    def hosts(self):
        hosts = {}
        for host in self.host_model.objects.filter(is_active=True):
            try:
                hosts.update({str(host): self.resource})
            except RequestException:
                pass
            except JSONDecodeError:
                pass
        return hosts

    def get_api_token(self, username):
        try:
            api_token = Token.objects.get(user__username=username).key
        except ObjectDoesNotExist:
            api_token = None
        return api_token

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            api_token=self.get_api_token(self.request.user),
            hosts=json.dumps(self.hosts),
            django_collect_offline_role=self.device_role,
            resource=self.resource,
        )
        return context



from django.db import models
from django.urls import reverse
from django.utils.safestring import mark_safe
from django_crypto_fields.constants import LOCAL_MODE
from django_crypto_fields.cryptor import Cryptor

from ..choices import ACTIONS


class TransactionModelMixin(models.Model):

    """Abstract model class for Incoming and Outgoing transactions.
    """

    tx = models.BinaryField()

    tx_name = models.CharField(
        max_length=64)

    tx_pk = models.UUIDField(
        db_index=True)

    producer = models.CharField(
        max_length=200,
        db_index=True,
        help_text='Producer name')

    action = models.CharField(
        max_length=1,
        choices=ACTIONS)

    timestamp = models.CharField(
        max_length=50,
        db_index=True)

    consumed_datetime = models.DateTimeField(
        null=True,
        blank=True)

    consumer = models.CharField(
        max_length=200,
        null=True,
        blank=True)

    is_ignored = models.BooleanField(
        default=False,
    )

    is_error = models.BooleanField(
        default=False)

    error = models.TextField(
        max_length=1000,
        null=True,
        blank=True)

    prev_batch_id = models.CharField(
        max_length=100,
        null=True,
        blank=True)

    batch_id = models.CharField(
        max_length=100,
        null=True,
        blank=True)

    def __str__(self):
        return f'{self._meta.model_name}.{self.tx_name}.{self.id}.{self.action}'

    def aes_decrypt(self, cipher):
        cryptor = Cryptor()
        plaintext = cryptor.aes_decrypt(cipher, LOCAL_MODE)
        return plaintext

    def aes_encrypt(self, plaintext):
        cryptor = Cryptor()
        cipher = cryptor.aes_encrypt(plaintext, LOCAL_MODE)
        return cipher

    def view(self):
        url = reverse('django_collect_offline:render_url',
                      kwargs={
                          'model_name': self._meta.object_name.lower(),
                          'pk': str(self.pk)})
        ret = mark_safe(
            '<a href="{url}" class="add-another" id="add_id_report" '
            'onclick="return showAddAnotherPopup(this);"> '
            '<img src="/static/admin/img/icon_addlink.gif" '
            'width="10" height="10" alt="View"/></a>'.format(url=url))
        return ret
    view.allow_tags = True

    class Meta:
        abstract = True


class HostModelMixin(models.Model):

    """Abstract class for hosts (either client or server).
    """

    hostname = models.CharField(
        max_length=200,
        unique=True)

    port = models.IntegerField(
        default='80')

    api_name = models.CharField(
        max_length=15,
        default='v1')

    format = models.CharField(
        max_length=15,
        default='json')

    authentication = models.CharField(
        max_length=15,
        default='api_key')

    is_active = models.BooleanField(
        default=True)

    last_sync_datetime = models.DateTimeField(
        null=True,
        blank=True)

    last_sync_status = models.CharField(
        max_length=250,
        default='-',
        null=True,
        blank=True)

    comment = models.TextField(
        max_length=50,
        null=True,
        blank=True)

    def __str__(self):
        return f'{self.hostname}:{self.port}'

    def natural_key(self):
        return (self.hostname, self.port, )

    @property
    def url_template(self):
        return (
            f'http://{self.hostname}:{self.port}/django_collect_offline/'
            f'api/{self.api_name}/')

    @property
    def url(self):
        return self.url_template.format(
            hostname=self.hostname, port=self.port, api_name=self.api_name)

    class Meta:
        abstract = True



import socket

from rest_framework import status
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.views import APIView

from ..models import (
    OutgoingTransaction, IncomingTransaction)


class TransactionCountView(APIView):
    """
    A view that returns the count  of transactions.
    """
    renderer_classes = (JSONRenderer,)

    def get(self, request):
        outgoingtransaction_count = OutgoingTransaction.objects.filter(
            is_consumed_server=False).count()
        outgoingtransaction_middleman_count = OutgoingTransaction.objects.filter(
            is_consumed_server=False,
            is_consumed_middleman=False).count()
        incomingtransaction_count = IncomingTransaction.objects.filter(
            is_consumed=False, is_ignored=False).count()
        content = {'outgoingtransaction_count': outgoingtransaction_count,
                   'outgoingtransaction_middleman_count': outgoingtransaction_middleman_count,
                   'incomingtransaction_count': incomingtransaction_count,
                   'hostname': socket.gethostname()}
        return Response(content, status=status.HTTP_200_OK)



from rest_framework import viewsets
from rest_framework.authentication import TokenAuthentication
from rest_framework.decorators import (
    api_view, authentication_classes, permission_classes)
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.reverse import reverse

from ..models import (
    OutgoingTransaction, IncomingTransaction)
from ..serializers import (
    OutgoingTransactionSerializer, IncomingTransactionSerializer)


@api_view(['GET'])
@authentication_classes((TokenAuthentication,))
@permission_classes((IsAuthenticated,))
def api_root(request, format=None):
    return Response({
        'outgoingtransaction': reverse('outgoingtransaction-list',
                                       request=request, format=format),
        'incomingtransaction': reverse('outgoingtransaction-list',
                                       request=request, format=format)})


class OutgoingTransactionViewSet(viewsets.ModelViewSet):

    queryset = OutgoingTransaction.objects.all()
    serializer_class = OutgoingTransactionSerializer

    def filter_queryset(self, queryset):
        return self.queryset.filter(
            is_consumed_server=False).order_by('timestamp')


class IncomingTransactionViewSet(viewsets.ModelViewSet):

    queryset = IncomingTransaction.objects.all()
    serializer_class = IncomingTransactionSerializer

    def filter_queryset(self, queryset):
        return self.queryset.filter(
            is_consumed=False, is_ignored=False).order_by('timestamp')
